/*
 * modified for lecture by Fabian Fuchs (KIT, Karlsruhe) based on:
 * 
 Copyright (c) 2007, Distributed Computing Group (DCG)
                    ETH Zurich
                    Switzerland
                    dcg.ethz.ch

 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

 - Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the
   distribution.

 - Neither the name 'Sinalgo' nor the names of its contributors may be
   used to endorse or promote products derived from this software
   without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package projects.uebung4extension.nodes.nodeImplementations;


import java.awt.Color;
import java.awt.Graphics;
import java.util.LinkedList;

import javax.swing.JOptionPane;

import projects.defaultProject.nodes.timers.MessageTimer;
import projects.uebung2.nodes.messages.DefaultMessage;
import projects.uebung3.nodes.nodeImplementations.GreedyFacetRoutingNode.GRFPayload;
import sinalgo.configuration.Configuration;
import sinalgo.configuration.WrongConfigurationException;
import sinalgo.gui.helper.NodeSelectionHandler;
import sinalgo.gui.transformation.PositionTransformation;
import sinalgo.nodes.Node;
import sinalgo.nodes.Position;
import sinalgo.nodes.edges.Edge;
import sinalgo.nodes.messages.Inbox;
import sinalgo.nodes.messages.Message;
import sinalgo.runtime.GUIRuntime;
import sinalgo.runtime.Main;
import sinalgo.tools.Tools;
import sinalgo.tools.Tuple;




/**
 * A leader election node 
 */
public class MISNode extends Node {

	/* store the current round */
	private int round;

	/* add additional variables / flags here */
	
	/* current count of white neighbors (NWN)
	 * (actually, NWN+1, as the node itself is counted)
	 * reset and updated each round
	 */
	private int numWhiteNeighbors;
	
	/* maximum NWN (numWhiteNeighbors) received by green neighbor 
	 * only required if the nodes state is GREEN 
	 */
	private int maxNWNGreenNeighbor;
	
	/*
	 * store the NWN value of the last round for
	 * comparison with received values (tiebreak)
	 */
	private int prevNWN;
	
	/* ID of the maxNWN node
	 * use for tie breaking in the GREEN state 
	 */
	private int maxNWNID;
	
	/* Flag indicating that the node may be isolated
	 */
	private boolean nodeMayBeIsolated;
	
	/* List of MIS nodes in 3-hop distance
	 * (used only if node becomes BLACK)
	 */
	LinkedList<MISConnectionInfo> connectMISList;
	
	
	public class MISConnectionInfo {
		public int MISNodeID;
		public int distance;
		public Tuple<Integer,Integer> hops;
		public MISConnectionInfo(int id, int dist, Tuple<Integer,Integer> h) {
			MISNodeID = id;
			distance = dist;
			hops = h;
		}
	}
	
	/* available states */
	public static final int WHITE = 0;
	public static final int GREEN = 1;
	public static final int GRAY  = 2;
	public static final int BLACK = 3;
	public static final int BLUE  = 4;
	
	private int state;
	
	public int getStatus() {
		return state;
	}
	
	/* payload for message */
	public class MISPayload {
		public int fromID;
		public int state;
		public int nwn; //numWhiteNeighbors of the neighbor
		public MISPayload(int ID, int s, int numWhiteNeighbors) {
			fromID = ID;
			state = s;
			nwn = numWhiteNeighbors;
		}
	}
	
	private boolean trueWithProbability(double prob) {
		return (Tools.getRandomNumberGenerator().nextDouble() < prob);
	}
	
	
	public class ConnectMISPayload {
		private int fromID;
		private int distance;
		private Tuple<Integer,Integer> hops;
		public ConnectMISPayload(int ID) {
			fromID = ID;
			distance = 0;
			hops = new Tuple<Integer,Integer>(-1,-1);
		}
		public ConnectMISPayload(int ID,int dist, Tuple<Integer,Integer> h) {
			fromID = ID;
			distance = dist;
			hops = h;
		}
		public void nextHop(int ID) {
			distance++;
			if (distance == 1) {
				hops.first = ID;
			} else if (distance == 2) {
				hops.second = ID;
			}
		}
		public int getDistance() {
			return distance;
		}
		public int getOriginId() {
			return fromID;
		}
		public Tuple<Integer,Integer> getHops() {
			return hops;
		}
	}
	
	public class JoinCDSPayload {
		public Tuple<Integer,Integer> nodes;
		public JoinCDSPayload (Tuple<Integer,Integer> n) {
			nodes = n;
		}
	}
	
	@Override
	public void handleMessages(Inbox inbox) {
		while(inbox.hasNext()) {
			Message msg = inbox.next();
			if(msg instanceof DefaultMessage<?>) {
				if (((DefaultMessage<?>) msg).getLoad() instanceof MISPayload) {
					if (state != GRAY && state != BLACK && state != BLUE) {
						//unpack packet
						MISPayload packet = (MISPayload)((DefaultMessage<?>) msg).getLoad();
						if (packet.state == BLACK) {
							state = GRAY;
						}
						if (packet.state == WHITE) {
							numWhiteNeighbors++;
						}
						if (packet.state == GREEN) {
							if (packet.nwn > maxNWNGreenNeighbor || 
									(packet.nwn == maxNWNGreenNeighbor && packet.fromID > maxNWNID)) {
								maxNWNGreenNeighbor = packet.nwn;
								maxNWNID = packet.fromID;
							}
						}
					}
				} else if (((DefaultMessage<?>) msg).getLoad() instanceof JoinCDSPayload) {
					//join CDS, regardless of state
					JoinCDSPayload packet = (JoinCDSPayload)((DefaultMessage<?>) msg).getLoad();
					if (packet.nodes.first == ID) {
						//this was the first (or only) node on the path, join CDS 
						state = BLUE;
					} else if (packet.nodes.second == ID) {
						//this was the first node on the path, join CDS and forward msg
						state = BLUE;
						this.send(msg, Tools.getNodeByID(packet.nodes.first));
					} else {
						Tools.appendToOutput("" + ID + " wrong node got joinCDSPayload\n");
					}
				} else if (((DefaultMessage<?>) msg).getLoad() instanceof ConnectMISPayload) {
					//received packet that discovers connection paths, check if node could be on such a path
					ConnectMISPayload packet = (ConnectMISPayload)((DefaultMessage<?>) msg).getLoad();
					
					//is the packets origin already reached with a better distance than the one of the packet?
					boolean newBestDistance = true;
					if (packet.fromID == ID) newBestDistance = false;
					for (MISConnectionInfo currentEntry : this.connectMISList) {
						if (currentEntry.MISNodeID == packet.fromID) {
							if (packet.getDistance()+1 >= currentEntry.distance) {
								newBestDistance = false;
							} else {// else: current entry is better, ignore the one just received
								connectMISList.remove(currentEntry);
							}
						}
					}
					if (newBestDistance) {
						//the path of the packet connects a new MIS node using a shortest distance
						//-> create a new copy of the packet (including hops - be careful to not over-reuse the objects)!
						ConnectMISPayload np = new ConnectMISPayload(packet.fromID,packet.getDistance(),
									new Tuple<Integer,Integer>(packet.getHops().first,packet.getHops().second));
						np.nextHop(ID);
						
						if (np.getDistance() <= 3){
							//if distance is larger, we do not care anymore, as MIS gets connected anyways
							
							//add path of the packet to connectMISList (as it is a new best distance)
							connectMISList.add(new MISConnectionInfo(np.getOriginId(),np.getDistance(),np.getHops()));
							
							//if node is already in MIS, notify the nodes on the path to join the CDS
							//let only the higher ID node select connecting paths
							if (state == BLACK && ID > np.fromID) { 
								//Tools.appendToOutput("" + ID + "MIS node received ConnectMIS\n");
								sendJoinCDS(np.fromID,np.getDistance(),np.getHops());
							}
						}
						if (packet.getDistance() < 3) {
							//forward message only if strictly smaller
							this.broadcast(new DefaultMessage<ConnectMISPayload>(np)); 
						}
					}
				}
			}
		}
	}
	
	private void sendJoinCDS(int MISNodeID, int distance,Tuple<Integer,Integer> hops) {
		DefaultMessage<JoinCDSPayload> joinMsg = 
				new DefaultMessage<JoinCDSPayload>(new JoinCDSPayload(hops));
//		Tools.appendToOutput(""+ID + " " + " BLACK " + MISNodeID + " dist " +
//				distance + " (" + hops.first + "," + hops.second + ")\n");
		if (hops.second == -1) { 
			//this node is neighbor of MIS node, should not happen as dist > 1
//			Tools.appendToOutput(""+ID + " c.hops.second=" +hops.second + 
//					" and dist=" +distance +" to MIS node " + MISNodeID +"\n");
		} else if (hops.second == ID) { 
			//only one intermediate node
			this.send(joinMsg, Tools.getNodeByID(hops.first));
		} else { 
			//two intermediate nodes
			this.send(joinMsg, Tools.getNodeByID(hops.second));
		}
	}

	@Override
	public void preStep() {
		//executed before each 'step' (i.e. timers and handleMessages)
		//I would personally recommend to only use the postStep method for simplicity
	}
	
	@Override
	public void postStep() {
		//executed after each 'step' (i.e. timers and handleMessages)
		round++;
		
//		if (state == BLACK) {
//			for (MISConnectionInfo c : this.connectMISList) {
//				Tools.appendToOutput("" + ID + " MIS " + c.MISNodeID + " dist " + c.distance + " (" + c.hops.first + "," + c.hops.second + ")\n");
//			}
//		}

		DefaultMessage<MISPayload> msg;

		switch (state) {
		case WHITE:
			if (numWhiteNeighbors == 0) {
				//node is / may be isolated
				if (nodeMayBeIsolated) { 
					/* If a node has no white neighbors in two consecutive rounds,
					 * this implies that the node is isolated. -> Join MIS 
					 * (using state GREEN to be sure)
					 */
					state = GREEN;
				} else { 
					/* In the first round without white neighbors, those neighbors 
					 * may temporarily be green and become white again. -> Use a flag 
					 */
					nodeMayBeIsolated = true;
				}
			} else { //otherwise node is not isolated
				//become green with probability 1/2w
				if (trueWithProbability(1/(2*(double)numWhiteNeighbors))) {
					state = GREEN;
				}
			}
			//transmit message
			msg = new DefaultMessage<MISPayload>(new MISPayload(ID, state, numWhiteNeighbors));
			this.broadcast(msg);
			break;
		case GREEN:
			/* Compare to the NWN (numWhiteNeighbors) value of the last round, 
			 * as that was received from the neighbor
			 */
			if (prevNWN > maxNWNGreenNeighbor || 
					prevNWN == maxNWNGreenNeighbor && ID > maxNWNID) {
				//Tools.appendToOutput("" + ID + " becoming black " + prevNWN + ">=" + maxNWNGreenNeighbor + " maxNWNID " + maxNWNID + "\n");
				state = BLACK;
				/* make nodes on shortest paths to other (already known) 
				 * close MIS nodes join the CDS */
				for (MISConnectionInfo c : this.connectMISList) {
					if (c.distance > 1 && ID > c.MISNodeID) { 
						//distance=1 means direct neighbor, let only the higher node ID select CDS paths
						//send join messages to nodes on a shortest path
						sendJoinCDS(c.MISNodeID,c.distance,c.hops);
					}
				}
				//init search shortest paths to other (already known) close MIS nodes
				DefaultMessage<ConnectMISPayload> connectMsg = new DefaultMessage<ConnectMISPayload>(new ConnectMISPayload(ID));
				this.broadcast(connectMsg);
			} else {
				//lost competition, revert to white state
				state = WHITE;
			}
			//tell neigbors about status (BLACK/WHITE)
			msg = new DefaultMessage<MISPayload>(new MISPayload(ID, state, numWhiteNeighbors));
			this.broadcast(msg);
			break;
		case BLACK:
			//nothing more to do :)
			break;
		case GRAY: 
			//nothing more to do :)
			break;
		case BLUE: 
			//nothing more to do :)
			break;
		default: 
			break;
		}
		maxNWNGreenNeighbor = 0;
		maxNWNID = -1;
		prevNWN = numWhiteNeighbors;
		numWhiteNeighbors = 0;
	}

	@Override
	public void init() {
		// initialize the node / read configuration values
		round = 0;
		maxNWNGreenNeighbor = 0;
		numWhiteNeighbors = 0;
		prevNWN = 0;
		state = WHITE;
		connectMISList = new LinkedList<MISConnectionInfo>();
	}

	@Override
	public void neighborhoodChange() {

	}
		
	/* (non-Javadoc)
	 * @see sinalgo.nodes.Node#draw(java.awt.Graphics, sinalgo.gui.transformation.PositionTransformation, boolean)
	 */
	public void draw(Graphics g, PositionTransformation pt, boolean highlight) {
		//set the color of this node, this allows us to see how the most important maxIDs spread
		
		switch (state) {
		case WHITE:
			this.setColor(Color.WHITE);
			break;
		case GREEN:
			this.setColor(Color.GREEN);
			break;
		case BLACK:
			this.setColor(Color.BLACK);
			break;
		case GRAY:
			this.setColor(Color.LIGHT_GRAY);  
			break;
		case BLUE:
			this.setColor(Color.BLUE);  
			break;
		default:
			this.setColor(Color.RED);
			break;
		}
		
		String text = Integer.toString(this.ID);
		// draw the node as a circle with the text inside
		super.drawNodeAsDiskWithText(g, pt, highlight, text, 10, Color.YELLOW);
	}


	@Override
	public String toString() {
		return "Node " + this.ID + " nwn " + this.numWhiteNeighbors; 
	}

	@Override
	public void checkRequirements() throws WrongConfigurationException {
		//does configuration fit the nodes requirements? 
		
	}
}
